from .interpreter import CoreProcessor
from .scope_types import Module, Scope
from .core_types import DotPath, RawToken, Alias, StretchSkipline, StretchTerminate, Stack
import sys

def run_stretch(dot_path):
    processor = CoreProcessor()
    module = Module(processor, DotPath(processor, dot_path))
    processor.run_scope(module)
    



class PrecedenceGraph:
    def __init__(self):
        self.groups = {}
        self.lower = {}
        self.higher = {}
        self.nodes = []
        self.members = {}
        
    def copy(self):
        new = PrecedenceGraph()
        new.groups = {name: set(members) for name, members in self.groups.items()}

        new.lower = {name: set(lows) for name, lows in self.lower.items()}
        new.higher = {name: set(highs) for name, highs in self.higher.items()}

        new.nodes = list(self.nodes)

        new.members = {name: set(members) for name, members in self.members.items()}

        return new

    
    def update(self, other):
        self.groups = {name: set(members) for name, members in other.groups.items()}

        self.lower = {name: set(lows) for name, lows in other.lower.items()}
        self.higher = {name: set(highs) for name, highs in other.higher.items()}

        self.nodes.extend(other.nodes)

        self.members.update({name: set(members) for name, members in other.members.items()})
        
    
    def group(self, name, low=None, high=None):
        if name not in self.groups:
            group = set()
            self.groups[name] = group
        else:
            group = self.groups[name]
        self.add_node(name)
        if low is not None: # group that is lower than me
            self.add_sub(name, low)
        if high is not None: # group that is higher than me
            self.add_dom(name, high)
            
    def add_sub(self, high, low):
        if high not in self.lower:
            self.lower[high] = set()
        self.lower[high].add(low)
        if low not in self.higher:
            self.higher[low] = set()
        self.higher[low].add(high)
    
    def add_dom(self, low, high):
        if high not in self.lower:
            self.lower[high] = set()
        self.lower[high].add(low)
        if low not in self.higher:
            self.higher[low] = set()
        self.higher[low].add(high)
        
    
    def add_node(self, group_name):
        if group_name not in self.nodes:
            self.nodes.append(group_name)

    def add_members(self, group, *members):
        for member in members:
            self.groups[group].add(member)
            
    def sort_group_names(self):
        order = []
        for group in self.nodes:
            pos = 0
            while pos < len(order):
                before = order[:pos]
                after = order[pos:]
                if any(lower in after for lower in self.lower.get(group, [])):
                    pos += 1
                    continue
                if any(higher in before for higher in self.higher.get(group, [])):
                    pos += 1
                    continue
                break
            order.insert(pos, group)
        
        return order
    
    def sort_groups(self):
        groups = []
        for name in self.sort_group_names():
            groups.append(self.groups[name])
        return list(reversed(groups))
        
        
                
        


from .core import __operators__ as core_operators, __terms__ as core_terms, __operator_precedence__ as core_precedence
from .core_types import Stack, DotPath, RawToken, StretchTerminate, StretchSkipline
from .scope_types import Scope, Module
from typing import Generator

class CoreProcessor:
    def __init__(self):
        self.__modules__ = {}
        self.__stack__ = Stack()
        self.running = True
        
        self.__operators__ = core_operators
        self.__terms__ = core_terms
        self.__operator_precedence__ = core_precedence

    def add_module(self, path, module:Module):
        if not isinstance(module, Module):
            raise TypeError(f"processor.add_module() expected a module, not {module.__class__.__name__} '{module}'")
        self.__modules__[path] = module
    
    def modules(self, name=None):
        if name is None:
            return self.__modules__
        return self.__modules__[name]
    
    # STACK
    def stack(self):
        return self.__stack__
    
    def process_line(self, scope:Scope, line):
        try:
            if line.startswith('<'):
                name, line = line.split(':')[:2]
                scope.store(RawToken(name[1:].strip()), line.strip())
            
            else:
                if line.startswith("try"):
                    try:
                        tokens = self.tokenize_line(scope, line[3:])
                        tokens = Stack(*tokens, single_layer=True)
                        self.apply_operators(scope, tokens) # apply scope / module level extension operators first
                        self.apply_operators(scope, tokens) # apply builtin operators
                
                        self.apply_terms(scope, tokens)
                        self.apply_terms(scope, tokens)
            
                        self.save_vars(scope, tokens)
                    except StretchSkipline:
                        pass
                    except StretchTerminate as e:
                        self.stack().push(e)
                    except Exception as e:
                        self.stack().push(StretchTerminate(f"{e}"))
                    
                else:
                    tokens = self.tokenize_line(scope, line)
                    tokens = Stack(*tokens, single_layer=True)
                    
                    self.apply_operators(scope, tokens) # apply scope / module level extension operators first
                    
                    self.apply_terms(scope, tokens)
                
                    self.save_vars(scope, tokens)
        except StretchSkipline:
            pass
            
    def apply_operators(self, scope, tokens):
        operators = scope.operators
        precedence = scope.precedence
        for precedence_group in precedence:
            if any([operator in tokens for operator in precedence_group]):
                index = 0
                while index < len(tokens):
                    if tokens[index] in precedence_group:
                        index -= 1
                        if not index >= 0:
                            try:
                                index += 1
                                op = tokens.pull(index)
                                r = tokens.pull(index)
                                tokens.insert(operators[op](self, scope, r), index)
                                continue
                            except TypeError as e:
                                msg = str(e)
                                if "missing" in msg or "positional argument" in msg:
                                    raise StretchTerminate(f"binary operator '{op.literal}' expected a left side operator")
                                raise StretchTerminate(e)
                            
                        l = tokens.pull(index)
                        op = tokens.pull(index)
                        r = tokens.pull(index)
                        tokens.insert(operators[op](self, scope, l, r), index)
                        continue
                    
                    index += 1
    def apply_terms(self, scope, tokens:Stack): # terms don't have precedence, they are applied right to left on the tokens to their right
        terms = scope.terms
        index = len(tokens) - 1
        while index >= 0:
            token = tokens.peek_if(RawToken, index)
            if token is not None:
                if token in terms:
                    tokens.pull(index)
                    r_toks = tokens[index:]
                    del tokens[index:]
                    
                    terms[token](self, scope, r_toks)
                    
                    tokens.extend(r_toks)
            index -= 1

            
    def save_vars(self, scope, tokens):
        index = len(tokens) - 1
        while index >= 0:
            identifier = tokens.pull_if((RawToken, DotPath), index)
            if identifier is not None:
                if isinstance(identifier, DotPath):
                    name = identifier.end().literal
                    
                    if name.endswith(':'):
                        name = name[:-1]
                        store_in = identifier.scope
                        if store_in is not None:
                            val = tokens.pull(index)
                            store_in.store(name, val)
                        
                elif isinstance(identifier, RawToken):
                    name = identifier.literal
                    if name.endswith(':'):
                        name = name[:-1]
                    name = RawToken(name)
                    if name in scope.operators:
                        raise StretchTerminate(f"Cannot store value in operator {identifier.literal}")
                    elif name in scope.terms:
                        raise StretchTerminate(f"Cannot store value in operator {identifier.literal}")
                    if index > len(tokens):
                        raise StretchTerminate(f"No value to store in {identifier.literal}")
                    val = tokens.pull(index)
                    scope.store(name, val)
            index -= 1
        
    def tokenize_line(self, scope, line):
        tokens = []
        tok_str = []
        start = None
        for chunk in line.split():
            
            if start is None:
                if chunk.startswith('"') and not chunk.endswith('"'):
                    start = '"'
            if start is not None:
                tok_str.append(chunk)
                if chunk.endswith(start):
                    tokens.append(" ".join(tok_str)[1:-1])
                    start = None
            else:
                for tok in self.parse_token(scope, chunk):
                    tokens.append(tok)
        return tokens
      
    def parse_token(self, scope, token:str):
        match token:
            case token if token.isdigit():
                yield int(token)
            case token if token.startswith('-') and token[1:].isdigit():
                yield int(token)
            case token if token.startswith('"') and token.endswith('"'):
                yield token.strip('"')
            case token if token.startswith("'") and token.endswith("'"):
                yield token.strip("'")
            case token if '.' in token:
                split = token.split('.')
                if len(split) == 2 and all([tok.strip('-').isdigit() for tok in split]):
                    yield float(token)
                else:
                    path = DotPath(*split, scope=scope)
                    if len(path.chain) == 1:
                        if not isinstance(path.chain[0], RawToken):
                            yield path.chain.pop()
                        else:
                            yield path
                            
                    else:
                        yield path
                        
            case token:
                raw_token = RawToken(token)
                if scope.contains(raw_token):
                    
                    yield scope.retrieve(raw_token)
                else:
                    yield raw_token
    
    def run_scope(self, scope:Scope):
        self.init_scope(scope)
        self.process_scope(scope)
    
    def init_scope(self, scope:Scope):
        scope.go_to(0)
        
        while scope.current_line < len(scope.__lines__):
            if not self.running:
                break
            
            line = scope.__lines__[scope.current_line]
            while line.endswith(':/'):
                scope.current_line += 1
                line = line[:-2] + " " + scope.__lines__[scope.current_line]

            if line.startswith("$"):
                self.process_line(scope, line[1:])
            scope.current_line += 1
        
    def process_scope(self, scope:Scope, start_line=0):
        
        scope.go_to(start_line)
        
        while scope.current_line < len(scope.__lines__):
            if not self.running:
                break
            line = scope.__lines__[scope.current_line]
            while line.endswith(':/'):
                scope.current_line += 1
                line = line[:-2] + " " + scope.__lines__[scope.current_line]

            if not line.startswith("$"):
                self.process_line(scope, line)
            scope.current_line += 1
    




from garnish import garnish
from .core_types import DotPath, RawToken, StretchTerminate, Alias, StretchSkipline, Stack, Loop
from .scope_types import Scope, Module
import importlib
from .precedence import PrecedenceGraph
from . import builtin_extensions
from typing import Any

__operators__ = {}
__operator_precedence__ = PrecedenceGraph()
__terms__ = {}

def add_group(name, low=None, high=None):
    __operator_precedence__.group(name, low, high)

@garnish
def operator(func, group:str, symbol:str):
    __operators__[RawToken(symbol)] = func
    __operator_precedence__.add_members(group, RawToken(symbol))
    return func
@garnish
def term(func, term:str):
    __terms__[RawToken(term)] = func
    return func

add_group('assignment', high="addition")
add_group('equivalence', 'assignment')

#OPERATORS
@operator.use('assignment','.')
def chain(proc, scope, l, r):
    if isinstance(l, RawToken):
        l = DotPath(l, scope=scope)
    elif isinstance(l, Module):
        return l.retrieve(r)
    elif isinstance(l, DotPath):
        l.extend(r, scope=scope)
        return l
    else:
        return getattr(l, r.literal)

@operator.use('equivalence', '==')
def is_eq(proc, scope, l, r):
    return l == r
@operator.use('equivalence', '!=')
def not_eq(proc, scope, l, r):
    return l != r
@operator.use('equivalence', '>')
def is_gt(proc, scope, l, r):
    return l > r
@operator.use('equivalence', '<')
def is_lt(proc, scope, l, r):
    return l < r


#TERMS
# error handling

@term.use("shutdown")
def shutdown(proc, scope, tokens):
    proc.running = False
    if len(tokens) > 0:
        raise StretchTerminate(tokens.pop())
    raise StretchTerminate()

@term.use("raise")
def error(proc, scope, tokens):
    message = tokens.pull_if(str)
    if message is None:
        message = "Exception without an error message"
    raise StretchTerminate(message)

@term.use('as')
def alias(proc, scope, tokens):
    tokens.insert(Alias(tokens.pull_only(RawToken)))
    
@term.use("stack")
def get_stack(proc, scope, tokens):
    tokens.insert(list(proc.stack()))

@term.use("display")
def display(proc, scope, tokens):
    print('\x1b[97m>\x1b[0m ' + ', '.join([str(tok).encode("utf-8").decode("unicode_escape") for tok in tokens]) + ' \x1b[97m<\x1b[0m')
    tokens.clear()
@term.use("print")
def _print(proc, scope, tokens):
    item = str(tokens.pull_if(object))
    if item is None:
        item = ""
    print(item.encode("utf-8").decode("unicode_escape"))

# control scope
@term.use("import")
def import_module(proc, scope, tokens):
    dotpath = tokens.pull_only((DotPath, RawToken), 0)
    if isinstance(dotpath, RawToken):
        dotpath = DotPath(dotpath, scope)
    path = dotpath.path_chain()
    if path in proc.modules():
        module = proc.modules(path)
    else:
        module = Module(proc, dotpath)
    proc.add_module(path, module)
    name = tokens.pull_if(Alias, 0)
    if name is not None:
        scope.store(name.name, module)
    tokens.insert(module)
    
@term.use("extend")
def import_extension(proc, scope, tokens):
    dotpath = tokens.pull_only((DotPath, RawToken))
    if isinstance(dotpath, RawToken):
        dotpath = DotPath(dotpath, scope)
    name = dotpath.file_name()
    if name.startswith('@'):
        extension = getattr(builtin_extensions, name[1:])
    else:
        extension = __import__(name)
    if hasattr(extension, '__extensions__'):
        extension.__extensions__.extend(proc, scope)

@term.use("init")
def init_scope(proc, scope, tokens):
    module = tokens.pull_only(Module)
    proc.init_scope(module)
    tokens.insert(module)
    
@term.use("enter")
def enter_scope(proc, scope, tokens):
    into = tokens.pull_only(Module, 0)
    line = tokens.pull_if(int, 0)
    if line is None:
        line = 0
    else:
        line -= 1
    if into is scope:
        
        into.go_to(line)
        raise StretchSkipline()
    else:
        proc.process_scope(into, line)
          
# control-flow  

@term.use("Line")
def get_line(proc, scope, tokens):
    tokens.insert(scope.current_line + 1)

@term.use("True")
def ret_true(proc, scope, tokens):
    tokens.insert(True)

@term.use("False")
def ret_false(proc, scope, tokens):
    tokens.insert(False)
@term.use("if")
def if_stat(proc, scope, tokens):
    token = tokens.pull_only(bool)
    if not token:
        raise StretchSkipline()
@term.use("unless")
def unless_stat(proc, scope, tokens):
    token = tokens.pull_only(bool)
    if token:
        raise StretchSkipline()

@term.use("goto")
def goto(proc, scope, tokens:Stack):
    
    to = tokens.pull_if((int, Loop)) or 0
    if isinstance(to, Loop):
        to = to.start
    prev = scope.current_line + 1
    scope.current_line = to - 1
    tokens.insert(prev)

@term.use("loop")
def startloop(proc, scope, tokens:Stack):
    goal = tokens.pull_only(int)
    prev = scope.current_line
    tokens.insert(Loop(prev, goal))
@term.use("continue")
def endloop(proc, scope, tokens:Stack):
    loop : Loop = tokens.pull_only(Loop)
    loop.loop(proc, scope)
    
@term.use("break")
def breakloop(proc, scope, tokens:Stack):
    loop : Loop = tokens.pull_only(Loop)
    loop.count = loop.goal
    

@term.use("Scope")
def get_scope(proc, scope, tokens):
    tokens.insert(scope)

@term.use("end")
def end_program(proc, scope, tokens):
    proc.running = False
    raise StretchSkipline()

@term.use("catch")
def catch_error(proc, scope, tokens):
    error = proc.stack().pull_if(Exception)
    if error is None:
        raise StretchSkipline()
    else:
        tokens.push(error)


# stack
@term.use("process")
def call_line(proc, scope, tokens:Stack):
    proc.process_line(scope, tokens.pull_only(str))

@term.use("push")
def push_stack(proc, scope, tokens):
    proc.stack().push(tokens.pull())

@term.use("pull")
def pull_stack(proc, scope, tokens):
    tokens.insert(proc.stack().pull())
    
        
    
    


    


import sys
from .scope_types import Scope

__types__ = {}
from garnish import garnish
@garnish
def add_type(type, determine=None):
    if determine is None:
        determine = type.__determine__
    __types__[determine] = type

class StretchTerminate(Exception):
    def __init__(self, *objects):
        message = ""
        for object in objects:
            message += str(object).encode("utf-8").decode("unicode_escape")
        super().__init__(("\x1b[31m" + message + "\x1b[0m"))
class StretchSkipline(Exception): pass

class Stack(list):
    __slots__ = ("single_layer")
    class StackError(StretchTerminate): pass
    def __init__(self, *iterable, single_layer=False):
        if single_layer:
            if any([isinstance(obj, Stack) and obj is not self for obj in iterable]):
                raise Stack.StackError(f"Cannot move vertically in a single layer stack")
        self.single_layer = single_layer
        super().__init__(iterable)
        
    def __repr__(self):
        return f"Stack -> {list.__repr__(self)}"    
    
    def level(self, index = 1):
        if self.single_layer:
            if index != 1:
                raise Stack.StackError(f"Cannot move vertically in a single layer stack")
            return self
        current_stack = self
        stacks = [current_stack]
        while len(current_stack) > 0:
            if isinstance(list.__getitem__(current_stack, -1), Stack):
                current_stack = current_stack[-1]
                stacks.append(current_stack)
            else:
                break
        return stacks[-index]
    
    def __getitem__(self, key):
        if isinstance(key, int):
            return list.__getitem__(self.level(), key)
        elif isinstance(key, slice):
            return Stack(*list.__getitem__(self.level(), key), single_layer=self.single_layer)
        
    
    def up(self, levels:int=1):
        if self.single_layer:
            raise Stack.StackError(f"Cannot move vertically in a single layer stack")
        for _ in range(levels):
            self.level().push(Stack())
    def down(self, levels:int = 1):
        if self.single_layer:
            raise Stack.StackError(f"Cannot move vertically in a single layer stack")
        return self.level(1 + levels).pop()
    
    def push(self, value):
        if self.single_layer:
            if isinstance(value, Stack) and value is not self:
                raise Stack.StackError(f"Cannot move vertically in a single layer stack")
        self.level().append(value)
    def pull(self, index=0):
        return self.level().pop(index)

    def pull_if(self, type, index=0):
        val = self.peek(index)
        if isinstance(val, type):
            self.pull(index)
            return val
        else:
            return None
    def pull_only(self, type, index=0):
        val = self.pull_if(type, index)
        if val is None:
            raise Stack.StackError(f"expected {type} at {index} on stack recieved {self.peek(index)}")
        return val
    
    
    def peek(self, index=0):
        if index > len(self.level()) - 1:
            return None
        return self.level()[index]

    def peek_if(self, type:type, index):
        val = self.peek(index)
        if not isinstance(val, type):
            return None
        return val
    def empty(self, level:int=1):
        self.level(level).clear()
    def copy(self, index = 0):
        self.push(self.peek(index))
    def surface(self, index=-1):
        self.push(self.pull(index))

    def insert(self, value, index=0):
        list.insert(self.level(), index, value)

class RawToken:
    __slots__ = ("literal",)
    __raw__ = {}
    
    def __new__(cls, literal:str, *args):
        if not isinstance(literal, str):
            raise StretchTerminate(f"Cannot create a RawToken from {literal}")
        if literal in cls.__raw__:
            instance = cls.__raw__[literal]
        else:
            instance = super().__new__(cls)
            instance.literal = literal
            cls.__raw__[literal] = instance
        return instance
    
    def __hash__(self):
        return hash((RawToken, self.literal))
    
    def __eq__(self, other):
        if isinstance(other, RawToken):
            return self.literal == other.literal
        elif isinstance(other, str):
            return False
        return False
    
    def __repr__(self):
        return f"(Raw:'{self.literal}')"
    
class DotPath:
    __slots__ = ("chain", "scope")
    def __init__(self, *chain, scope=None):
        self.scope = scope
        self.chain = []
        for segment in chain:
            if isinstance(segment, str):
                for str_seg in segment.strip().split('.'):
                    identifier = RawToken(str_seg)
                    
                    if scope is not None:
                        if scope.contains(identifier):
                            val = scope.retrieve(identifier)
                            if isinstance(val, Scope):
                                scope = val
                                self.scope = val
                            else:
                                
                                self.chain.append(val)
                        else:
                            
                            self.chain.append(identifier)
                                
                    else:
                        self.chain.append(identifier)
            elif isinstance(segment, RawToken):
                self.chain.append(segment)
    
    def __repr__(self):
        return f"<{'.'.join([str(seg) if not isinstance(seg, RawToken) else seg.literal for seg in self.chain])}>"
    
    def extend(self, *segments, scope=None):
        for segment in segments:
            if isinstance(segment, str):
                for str_seg in segment.strip().split('.'):
                    if scope is not None:
                        self.chain.append(scope.parse_raw(RawToken(str_seg)))
                    else:
                        self.chain.append(RawToken(str_seg))
            elif isinstance(segment, RawToken):
                self.chain.append(segment)
    
    def file_name(self, ext:str=None) -> str:
        if ext is not None: ext = '.' + ext
        return "/".join([name.literal for name in self.chain]) + (ext or "")
    def path_chain(self):
        return ".".join([name.literal for name in self.chain])
    
    def end(self) -> RawToken:
        return self.chain[-1]

class Loop:
    __slots__ = ("scope", "start", "count", "goal")
    def __init__(self, start_line, goal):
        self.start = start_line
        self.goal = goal
        self.count = 0
    
    def loop(self, proc, scope):
        self.count += 1
        if self.count < self.goal:
            scope.current_line = self.start
            raise StretchSkipline()
        

class Alias:
    __slots__ = ("name")
    
    def __init__(self, name):
        self.name = name




type DotPath = None #alias

class Scope:
    def __init__(self, processor, name, parent):
        self.processor = processor
        self.name = name
        self.parent = parent
        
        self.__lines__ = []
        self.current_line = 0
        self.__scope__ = {}
    
    def store(self, name, value):
        self.__scope__[name] = value
    
    def retrieve(self, name):
        return self.__scope__[name]
    
    def swap(self, name, value):
        temp = self.__scope__[name]
        self.__scope__[name] = value
        return temp
    
    def erase(self, name):
        del self.__scope__[name]
    
    def contains(self, name):
        return name in self.__scope__
    
    def clear_line(self, line:int):
        del self.__lines__[line]
    def clear_lines(self):
        self.__lines__.clear()
    def write_line(self, line:str):
        self.__lines__.append(line.strip())
    def write_lines(self, lines:list[str]):
        for line in lines:
            self.write_line(line)
    
    def operators(self, symbol=None):
        return self.parent.operators(symbol)
    def terms(self, term=None):
        return self.parent.terms(term)

    def __iter__(self):
        while self.current_line < len(self.__lines__):
            yield self.__lines__[self.current_line]
            self.current_line += 1
    
    def go_to(self, line):
        self.current_line = line

class Module(Scope):
    @staticmethod
    def load_module(file_path):
        with open(file_path, 'r') as source:
            return source.readlines()
    
    def __new__(cls, processor, module_path:DotPath, parent=None):
        path = module_path.path_chain()
        if path in processor.modules():
            instance = processor.modules(path)
        else:
            instance = super().__new__(cls)
            processor.add_module(path, instance)
        return instance
    
    def register_operator(self, symbol:str, operation):
        if symbol not in self.__operators__:
            self.__operators__[symbol] = operation
        
    def __init__(self, processor, module_path:DotPath, parent=None):
        super().__init__(processor, module_path.file_name(), parent)
        self.path = module_path.file_name('str')
        self.write_lines(self.load_module(self.path))
        
        
        self.__operators__ = processor.__operators__
        self.__operator_precedence__ = processor.__operator_precedence__
        self.__terms__ = processor.__terms__
        self.__types__ = {}
        self.__scope__ = {}
    @property
    def operators(self):
        return self.__operators__
    @property
    def precedence(self):
        return self.__operator_precedence__.sort_groups()
    
    @property
    def terms(self):
        return self.__terms__


import importlib
from stretch.core_types import StretchTerminate
imported = {}

def __getattr__(extension_name):
    if extension_name in imported:
        extension = imported[extension_name]
    else:
        try:
            extension = importlib.import_module("." + extension_name, 'stretch.builtin_extensions')
        except:
            raise StretchTerminate(f"Stretch builtin extensions does not contain '{extension_name}'")
        imported[extension_name] = extension
    return extension




from garnish import garnish
from .core import __operators__, __terms__, __operator_precedence__
from .core_types import RawToken

class Extender:
    def __init__(self):
        self.__operators__ = __operators__.copy()
        self.__operator_precedence__ = __operator_precedence__.copy()
        self.__terms__ = __terms__.copy()
        
        self.__direct__ = []
    
    
    def extend(self, proc, scope):
        scope.__operators__.update(self.__operators__)
        scope.__operator_precedence__.update(self.__operator_precedence__)
        scope.__terms__.update(self.__terms__)
        for func in self.__direct__:
            func(proc, scope)
    
    
    def __enter__(self):
        return self
    def __exit__(self, *exc_args): pass
    
    def __call__(self, func):
        self.__direct__.append(func)
    
    def add_operator_group(self, name, low=None, high=None):
        self.__operator_precedence__.group(name, low, high)
    
    @garnish
    def operator(self, func, group:str, symbol:str):
        symbol = RawToken(symbol)
        self.__operators__[symbol] = func
        self.__operator_precedence__.add_members(group, symbol)
        return func

    @garnish
    def term(self, func, term:str):
        self.__terms__[RawToken(term)] = func
        return func






from stretch.extension import Extender

__extensions__ = Extender()
with __extensions__ as extend:
    @extend.term.use("in")
    def get_input(proc, scope, tokens):
        message = tokens.pull_if(str)
        if message is None:
            message = ""
        message = message.encode("utf-8").decode("unicode_escape")  
        tokens.insert(input(message))
    @extend.term.use("ins")
    def get_input(proc, scope, tokens):
        message = tokens.pull_if(str)
        if message is None:
            message = ""
        message = message.encode("utf-8").decode("unicode_escape")
        user_input = "/"
        while user_input.endswith("/"):
            user_input = user_input[:-1]
            user_input += input(message)
        tokens.insert(user_input)
    


from stretch.extension import Extender
import math

__extensions__ = Extender()
with __extensions__ as extend:
    
    extend.add_operator_group("exponentiation", "multiplication")
    extend.add_operator_group("multiplication", "addition")
    extend.add_operator_group("addition", "equivalence")
    
    @extend.operator.use("exponentiation", "^")
    def __pow__(proc, scope, l, r):
        return math.pow(l, r)
    
    @extend.operator.use("multiplication", "*")
    def __mul__(proc, scope, l, r):
        return l * r
    
    @extend.operator.use("multiplication", "/")
    def __div__(proc, scope, l, r):
        return l / r
    
    @extend.operator.use("addition", "+")
    def __add__(proc, scope, l, r):
        return l + r

    @extend.operator.use("addition", "-")
    def __subtract__(proc, scope, l, r=None):
        if r is None:
            return -l
        return l - r
    
    @extend.term.use("int")
    def make_int(proc, scope, tokens):
        tokens.insert(int(tokens.pull()))
    
    @extend.term.use("float")
    def make_float(proc, scope, tokens):
        tokens.insert(float(tokens.pull()))
    